import sqlite3

from typing import Any, Literal, Optional, Type, TypeVar, Callable, Union
from ._connection import Connection
from ._util import _exec

connect = Connection 

async def complete_statement(statement: str) -> bool:
    return await _exec(sqlite3.complete_statement, statement)

async def enable_callback_tracebacks(flag: bool,) -> None:
    return await _exec(sqlite3.enable_callback_tracebacks, flag)

AdapterType = TypeVar('AdapterType')

async def register_adapter(
    type: Type[AdapterType],
    adapter: Callable[[AdapterType], Optional[Union[int, float, str, bytes]]],
) -> None:
    return await _exec(sqlite3.register_adapter, type, adapter)

async def register_converter(
    typename: Literal['NULL', 'INTEGER', 'REAL', 'TEXT', 'BLOB'],
    converter: Callable[[bytes], Any]
) -> None:
    return await _exec(sqlite3.register_converter, typename, converter)



from collections import deque
from sys import version_info
from typing import Optional, Type
if version_info >= (3, 11):
    from typing import SupportsIndex
    from collections.abc import Coroutine
    from typing import NoReturn
    import sqlite3
    from types import TracebackType
    from . import _connection

    class Blob:
        __slots__ = (
            '__connection',
            '__blob',
            '__setitems',
            '__weakref__',
            '__dict__',
        )

        def __init__(
            self,
            connection: '_connection.Connection',
            blob: 'sqlite3.Blob',
        ) -> None:
            self.__connection = connection
            self.__blob = blob
            self.__setitems: deque[Coroutine[None, None, None]] = deque()

        async def __aenter__(self) -> 'Blob':
            return self

        async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc: Optional[BaseException],
            traceback: Optional[TracebackType],
        ) -> None:
            await self.close()

        async def close(self) -> None:
            try:
                await self.flush()
            finally:
                await self.__connection._exec(self.__blob.close)

        async def read(self, length: int = -1) -> bytes:
            await self.flush()
            return await self.__connection._exec(self.__blob.read, length)

        async def write(self, data: bytes) -> None:
            await self.flush()
            await self.__connection._exec(self.__blob.write, bytes)

        async def tell(self) -> int:
            await self.flush()
            return await self.__connection._exec(self.__blob.tell)

        async def seek(self, offset: int, origin: int = 0) -> None:
            await self.flush()
            await self.__connection._exec(self.__blob.seek, offset, origin)

        async def flush(self) -> None:
            while self.__setitems:
                await self.__setitems.popleft()

        async def __getitiem__(self, index: SupportsIndex | slice) -> int | bytes:
            await self.flush()
            return await self.__connection._exec(self.__blob.__getitem__, index)

        async def set(self, index: SupportsIndex | slice, value: int | bytes) -> None:
            await self.flush()
            await self.__set(index, value)

        async def __set(self, index: SupportsIndex | slice, value: int | bytes) -> None:
            await self.__connection._exec(self.__blob.__setitem__, index, value)

        def __setitem__(self, index: SupportsIndex | slice, value: int | bytes) -> None:
            self.__setitems.append(self.__set(index, value))

        def __len__(self) -> NoReturn:
            raise NotImplementedError('Use the len method instead')

        def __bool__(self) -> NoReturn:
            raise NotImplementedError('Use the bool method instead')

        async def len(self) -> int:
            await self.flush()
            return await self.__connection._exec(len, self.__blob)

        async def bool(self) -> bool:
            await self.flush()
            return await self.__connection._exec(bool, self.__blob)





from functools import partial
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import Any, AsyncIterable, Iterator, AsyncIterator, cast, Callable
from asyncio import get_running_loop
from . import _connection
from ._types import ExecReturn

class _Unset:
    __slots__ = ()

_unset = _Unset()

def _close_in_executor(executor: Executor, closeable: Any):
    '''A blocking finalizer that runs in the given executor.

    This should ideally never be relied upon.
    '''
    executor.submit(closeable.close).result()
    executor.shutdown()

_executor = ThreadPoolExecutor(max_workers=1)

async def _exec(
    function: Callable[..., ExecReturn],
    *args: Any,
    **kwargs: Any,
) -> ExecReturn:
    return await get_running_loop().run_in_executor(
        _executor,
        partial(function, *args, **kwargs),
    )

class _IterDump(AsyncIterable[str]):
    __slots__ = (
        '__connection',
        '__iterdump',
    )

    def __init__(
        self,
        connection: '_connection.Connection',
        iterdump: Callable[[], Iterator[str]],
    ) -> None:
        self.__connection = connection
        self.__iterdump = iterdump

    async def __aiter__(self) -> AsyncIterator[str]:
        iterator = await self.__connection._exec(self.__iterdump)
        return _IterDumpIterator(self.__connection, iterator)

class _IterDumpIterator(AsyncIterable[str]):
    __slots__ = (
        '__connection',
        '__iterator',
    )

    def __init__(
        self,
        connection: '_connection.Connection',
        iterator: Iterator[str]
    ) -> None:
        self.__connection = connection
        self.__iterator

    async def __anext__(self) -> str:
        try:
            return await self.__connection._exec(
                cast(Callable[..., str], next),
                self.__iterator,
            )
        except StopIteration:
            raise StopAsyncIteration



'''Executor-base Asynchronous sqlite3 module wrapper.
'''
from ._blob import * # noqa
from ._connection import * # noqa
from ._cursor import * # noqa
from ._global import * # noqa
from ._types import * # noqa



import sqlite3

from typing import Any, Literal, TypeVar, Callable

ExecReturn = TypeVar('ExecReturn')
IsolationLevel = Literal['DEFERRED', 'IMMEDIATE', 'EXCLUSIVE']
RowFactory = Callable[[sqlite3.Cursor, tuple], Any]
TextFactory = Callable[[str], Any]



import sqlite3

from types import TracebackType
from typing import Any, Iterable, List, Optional, Tuple, Type, Union, overload
from . import _connection
from ._util import _Unset, _unset
from ._types import RowFactory
from asyncio import Lock

class _RaisedStopIteration:
    __slots__ = ()

_raised_stop_iteration = _RaisedStopIteration()

def _next(it) -> Any:
    '''Calls next, turning a raised stopiteration into _raised_stop_iteration.

    We need this because coroutines are not allowed to throw StopIteration at
    all.  They get converted to RuntimeError.
    '''
    try:
        return next(it)
    except StopIteration:
        return _raised_stop_iteration

class Cursor:
    __slots__ = (
        '__connection',
        '__factory',
        '__cursor',
        '__opened',
        '__lock',
        '__weakref__',
        '__dict__',
    )

    def __init__(
        self,
        connection: '_connection.Connection',
        factory: Type[sqlite3.Cursor] = sqlite3.Cursor,
    ) -> None:
        self.__connection = connection
        self.__factory = factory
        self.__opened = False

        # Lock for opening and closing.
        self.__lock = Lock()

    async def _open(self) -> None:
        async with self.__lock:
            if not self.__opened:
                await self.__connection._open()
                self.__cursor = await self.__connection._exec(
                    self.__connection._connection.cursor,
                    self.__factory,
                )
                self.__opened = True

    async def __aenter__(self) -> 'Cursor':
        await self._open()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def execute(self, sql: str, parameters: Iterable[Any] = ()) -> 'Cursor':
        await self._open()
        await self.__connection._exec(self.__cursor.execute, sql, parameters)
        return self

    async def executemany(self, sql: str, parameters: Iterable[Any]) -> 'Cursor':
        await self._open()
        await self.__connection._exec(self.__cursor.executemany, sql, parameters)
        return self

    async def executescript(self, sql_script: str) -> 'Cursor':
        await self._open()
        await self.__connection._exec(self.__cursor.executescript, sql_script)
        return self

    async def close(self) -> None:
        async with self.__lock:
            if self.__opened:
                try:
                    await self.__connection._exec(self.__cursor.close)
                finally:
                    self.__opened = False

    async def fetchone(self) -> Any:
        await self._open()
        return await self.__connection._exec(self.__cursor.fetchone)

    async def fetchmany(self, size: Optional[int] = None) -> List[Any]:
        await self._open()
        if size is not None:
            return await self.__connection._exec(self.__cursor.fetchmany, size)
        else:
            return await self.__connection._exec(self.__cursor.fetchmany)

    async def fetchall(self) -> List[Any]:
        await self._open()
        return await self.__connection._exec(self.__cursor.fetchall)

    async def setinputsizes(self, sizes: Any) -> None:
        await self._open()
        await self.__connection._exec(self.__cursor.setinputsizes, sizes)

    async def setoutputsize(self, size: Any, column: Any = None) -> None:
        await self._open()
        await self.__connection._exec(self.__cursor.setoutputsize, size, column)

    @overload
    async def arraysize(self) -> int:
        ...

    @overload
    async def arraysize(self, value: int) -> None:
        ...

    async def arraysize(self, value: Union[int, _Unset] = _unset) -> Optional[int]:
        await self._open()
        if value is _unset:
            return await self.__connection._exec(
                getattr,
                self.__cursor,
                'arraysize',
            )
        else:
            await self.__connection._exec(
                setattr,
                self.__cursor,
                'arraysize',
                value,
            )

    @property
    def connection(self) -> '_connection.Connection':
        return self.__connection

    async def description(
        self,
    ) -> Optional[Tuple[Tuple[str, None, None, None, None, None, None], ...]]:
        await self._open()
        return await self.__connection._exec(
            getattr,
            self.__cursor,
            'description',
        )

    async def lastrowid(self) -> Optional[int]:
        await self._open()
        return await self.__connection._exec(
            getattr,
            self.__cursor,
            'lastrowid',
        )

    async def rowcount(self) -> int:
        await self._open()
        return await self.__connection._exec(
            getattr,
            self.__cursor,
            'rowcount',
        )

    @overload
    async def row_factory(self) -> Optional[RowFactory]:
        ...

    @overload
    async def row_factory(self, value: Optional[RowFactory]) -> None:
        ...
            
    async def row_factory(
        self,
        value: Optional[Union[RowFactory, _Unset]] = _unset,
    ) -> Optional[RowFactory]:
        await self._open()
        if value is _unset:
            return await self.__connection._exec(
                getattr,
                self.__cursor,
                'row_factory',
        )
        else:
            await self.__connection._exec(
                setattr,
                self.__cursor,
                'row_factory',
                value,
        )

    def __aiter__(self) -> 'Cursor':
        return self

    async def __anext__(self) -> Any:
        await self._open()
        # We need this because coroutines are not allowed to throw StopIteration
        # at all.  They get converted to RuntimeError
        item = await self.__connection._exec(
            _next,
            self.__cursor,
        )
        if item is _raised_stop_iteration:
            raise StopAsyncIteration
        else:
            return item
        



from os import PathLike
import sqlite3

from functools import partial
from concurrent.futures import Executor, ThreadPoolExecutor
from types import TracebackType
from typing import (
    Any, AsyncIterable, Iterable, Optional, Type, Union, overload, Callable
)
from weakref import finalize
from asyncio import get_running_loop, Lock
from sys import version_info
from . import _blob, _cursor
from ._util import _Unset, _unset, _IterDump, _close_in_executor
from ._types import ExecReturn, IsolationLevel, RowFactory, TextFactory

class Connection:
    __slots__ = (
        '__cached_statements',
        '__check_same_thread',
        '_connection',
        '__database',
        '__detect_types',
        '__dict__',
        '__executor_arg',
        '__executor',
        '__factory',
        '__finalizer',
        '__isolation_level',
        '__lock',
        '__opened',
        '__timeout',
        '__uri',
        '__weakref__',
    )
    def __init__(
        self,
        database: Union[str, bytes, PathLike],
        timeout: float = 5.0,
        detect_types: int = 0,
        isolation_level: Optional[IsolationLevel] = 'DEFERRED',
        check_same_thread: bool = __debug__,
        factory: Type[sqlite3.Connection] = sqlite3.Connection,
        cached_statements: int = 128,
        uri: bool = False,
        executor: Optional[Executor] = None,
    ) -> None:
        self.__database = database
        self.__timeout = timeout
        self.__detect_types = detect_types
        self.__isolation_level = isolation_level
        self.__check_same_thread = check_same_thread
        self.__factory = factory
        self.__cached_statements = cached_statements
        self.__uri = uri
        self.__executor_arg = executor

        self.__opened = False

        # Lock for opening and closing.
        self.__lock = Lock()

    async def _exec(
        self,
        function: Callable[..., ExecReturn],
        *args: Any,
        **kwargs: Any,
    ) -> ExecReturn:
        return await get_running_loop().run_in_executor(
            self.__executor,
            partial(function, *args, **kwargs),
        )

    async def _open(self) -> None:
        '''Open the connection if it's not opened.
        '''
        async with self.__lock:
            if not self.__opened:
                if self.__executor_arg is None:
                    self.__executor = ThreadPoolExecutor(max_workers=1)
                else:
                    self.__executor = self.__executor_arg

                self._connection = await self._exec(
                    sqlite3.connect,
                    database = self.__database,
                    timeout = self.__timeout,
                    detect_types = self.__detect_types,
                    isolation_level = self.__isolation_level,
                    check_same_thread = self.__check_same_thread,
                    factory = self.__factory,
                    cached_statements = self.__cached_statements,
                    uri = self.__uri,
                )

                self.__finalizer = finalize(
                    self,
                    _close_in_executor,
                    self.__executor,
                    self._connection,
                )

                self.__opened = True


    async def __aenter__(self) -> 'Connection':
        await self._open()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        try:
            if exc is None:
                await self.commit()
            else:
                await self.rollback()
        finally:
            await self.close()

    def cursor(
        self,
        factory: Type[sqlite3.Cursor] = sqlite3.Cursor,
    ) -> '_cursor.Cursor':
        return _cursor.Cursor(self, factory)

    if version_info >= (3, 11):
        async def blobopen(
            self,
            table: str,
            column: str,
            row: str,
            *,
            readonly: bool = False,
            name: str = 'main',
        ) -> '_blob.Blob':
            await self._open()

            return _blob.Blob(
                self,
                await self._exec(self._connection.blobopen,
                    table,
                    column,
                    row,
                    readonly=readonly,
                    name=name,
                ),
            )

    async def commit(self) -> None:
        await self._exec(self._connection.commit)

    async def close(self) -> None:
        async with self.__lock:
            if self.__opened:
                try:
                    self.__opened = False
                    self.__finalizer.detach()
                    await self._exec(self._connection.close)
                    if self.__executor_arg is None:
                        self.__executor.shutdown()
                finally:
                    del self.__finalizer, self.__executor, self._connection

    async def rollback(self) -> None:
        await self._exec(self._connection.rollback)

    async def execute(
        self,
        sql: str,
        parameters: Iterable[Any] = (),
    ) -> '_cursor.Cursor':
        return await self.cursor().execute(sql, parameters)

    async def executemany(
        self,
        sql: str,
        parameters: Iterable[Any],
    ) -> '_cursor.Cursor':
        return await self.cursor().executemany(sql, parameters)

    async def executescript(self, sql_script: str) -> '_cursor.Cursor':
        return await self.cursor().executescript(sql_script)

    if version_info >= (3, 8):
        async def create_function( # type: ignore
            self,
            name: str,
            narg: int,
            func: Callable[..., Optional[Any]],
            *,
            deterministic: bool = False,
        ) -> None:
            await self._exec(
                self._connection.create_function,
                name,
                narg,
                func,
                deterministic=deterministic,
            )
    else:
        async def create_function(
            self,
            name: str,
            narg: int,
            func: Callable[..., Optional[Any]],
        ) -> None:
            await self._exec(
                self._connection.create_function,
                name,
                narg,
                func,
            )

    async def create_aggregate(
        self,
        name: str,
        n_arg: int,
        aggregate_class: Optional[type],
    ) -> None:
        await self._exec(
            self._connection.create_aggregate,
            name,
            n_arg,
            aggregate_class,
        )

    if version_info >= (3, 11):
        async def create_window_function(
            self,
            name: str,
            num_params: int,
            aggregate_class: Optional[type],
        ) -> None:
            await self._exec(
                self._connection.create_window_function,
                name,
                num_params,
                aggregate_class,
            )

    async def create_collation(
        self,
        name: str,
        callable: Callable[[str, str], Optional[int]],
    ) -> None:
        await self._exec(
            self._connection.create_collation,
            name,
            callable,
        )

    def interrupt(self) -> None:
        self._connection.interrupt()

    async def set_authorizer(
        self,
        authorizer_callback: Callable[[
            int,
            Optional[str],
            Optional[str],
            Optional[str],
            Optional[str],
        ], Optional[int]],
    ) -> None:
        await self._exec(
            self._connection.set_authorizer,
            authorizer_callback,
        )

    async def set_progress_handler(
        self,
        progress_handler: Callable[[], Optional[int]],
        n: int,
    ) -> None:
        await self._exec(
            self._connection.set_progress_handler,
            progress_handler,
            n,
        )
    async def set_trace_callback(
        self,
        trace_callback: Callable[[str], Optional[Any]],
    ) -> None:
        await self._exec(
            self._connection.set_trace_callback,
            trace_callback,
        )

    async def enable_load_extension(
        self,
        enabled: bool,
    ) -> None:
        await self._exec(
            self._connection.enable_load_extension,
            enabled,
        )

    async def load_extension(
        self,
        path: str,
    ) -> None:
        await self._exec(
            self._connection.load_extension,
            path,
        )

    def iterdump(
        self,
    ) -> AsyncIterable[str]:
        return _IterDump(self, self._connection.iterdump)

    async def backup(
        self,
        target: Union['Connection', sqlite3.Connection],
        *,
        pages: int = -1,
        progress: Callable[[int, int, int], Optional[Any]] = None,
        name: str = 'main',
        sleep: float = 0.250,
    ) -> None:
        await self._exec(
            self._connection.backup,
            target,
            pages=pages,
            progress=progress,
            name=name,
            sleep=sleep,
        )

    async def getlimit(self, category: int) -> int:
        return await self._exec(
            self._connection.getlimit,
            category,
        )
    async def setlimit(self, category: int, limit: int) -> int:
        return await self._exec(
            self._connection.setlimit,
            category,
            limit,
        )

    if version_info >= (3, 11):
        async def serialize(self, *, name: str = 'main') -> bytes:
            return await self._exec(
                self._connection.serialize,
                name=name,
            )

        async def deserialize(self, data: bytes, *, name: str = 'main') -> None:
            await self._exec(
                self._connection.deserialize,
                bytes,
                name=name,
            )

    
    async def in_transaction(self) -> bool:
        return await self._exec(getattr, self._connection, 'in_transaction')

    @overload
    async def isolation_level(self) -> Optional[IsolationLevel]:
        ...

    @overload
    async def isolation_level(self, value: Optional[IsolationLevel]) -> None:
        ...

    async def isolation_level(
        self,
        value: Optional[Union[IsolationLevel, _Unset]] = _unset,
    ) -> Optional[IsolationLevel]:
        if value is _unset:
            return await self._exec(getattr, self._connection, 'isolation_level')
        else:
            await self._exec(setattr, self._connection, 'isolation_level', value)

    @overload
    async def row_factory(self) -> Optional[RowFactory]:
        ...

    @overload
    async def row_factory(self, value: Optional[RowFactory]) -> None:
        ...
            
    async def row_factory(
        self,
        value: Optional[Union[RowFactory, _Unset]] = _unset,
    ) -> Optional[RowFactory]:
        if value is _unset:
            return await self._exec(getattr, self._connection, 'row_factory')
        else:
            await self._exec(setattr, self._connection, 'row_factory', value)

    @overload
    async def text_factory(self) -> TextFactory:
        ...

    @overload
    async def text_factory(self, value: TextFactory) -> None:
        ...
            
    async def text_factory(
        self,
        value: Union[TextFactory, _Unset] = _unset,
    ) -> Optional[TextFactory]:
        if value is _unset:
            return await self._exec(getattr, self._connection, 'text_factory')
        else:
            await self._exec(setattr, self._connection, 'text_factory', value)
            
    async def total_changes(self) -> int:
        return await self._exec(getattr, self._connection, 'total_changes')



import unittest
import easqlite
import tempfile
from pathlib import Path

class SimpleTests(unittest.IsolatedAsyncioTestCase):
    async def test_some_queries(self):
        with tempfile.TemporaryDirectory() as dir:
            Path(dir) / "test.db"
            async with easqlite.connect(
                Path(dir) / "test.db",
                isolation_level=None,
            ) as connection:
                await connection.execute('''
                    CREATE TABLE foo (
                        id INTEGER PRIMARY KEY NOT NULL,
                        alpha TEXT NULL,
                        beta BLOB NOT NULL
                    )
                ''')

                async with connection.cursor() as cursor:
                    sql = 'INSERT INTO foo (id, alpha, beta) VALUES (?, ?, ?)'
                    await cursor.execute(sql, (1, None, b'Beta test one'))
                    await cursor.execute(sql, (2, 'Alpha test two', b'Beta test two'))
                    await connection.commit()


                async with connection.cursor() as cursor:
                    sql = 'SELECT id, alpha, beta FROM foo ORDER BY id ASC'
                    await cursor.execute(sql)
                    self.assertEqual(
                        await cursor.fetchone(),
                        (1, None, b'Beta test one'),
                    )
                    self.assertEqual(
                        await cursor.fetchone(),
                        (2, 'Alpha test two', b'Beta test two'),
                    )
                    self.assertIs(await cursor.fetchone(), None)

                async with connection.cursor() as cursor:
                    sql = 'SELECT id, alpha, beta FROM foo ORDER BY id ASC'
                    await cursor.execute(sql)
                    self.assertEqual(
                        await cursor.fetchmany(),
                        [(1, None, b'Beta test one')],
                    )
                    self.assertEqual(
                        await cursor.fetchmany(),
                        [(2, 'Alpha test two', b'Beta test two')],
                    )
                    self.assertEqual(
                        await cursor.fetchmany(),
                        [],
                    )

                async with connection.cursor() as cursor:
                    sql = 'SELECT id, alpha, beta FROM foo ORDER BY id ASC'
                    await cursor.execute(sql)
                    self.assertEqual(
                        await cursor.fetchmany(1000),
                        [
                        (1, None, b'Beta test one'),
                        (2, 'Alpha test two', b'Beta test two'),
                    ])

                async with connection.cursor() as cursor:
                    sql = 'SELECT id, alpha, beta FROM foo ORDER BY id ASC'
                    await cursor.execute(sql)
                    await cursor.arraysize(-1)
                    self.assertEqual(
                        await cursor.fetchmany(),
                        [
                            (1, None, b'Beta test one'),
                            (2, 'Alpha test two', b'Beta test two'),
                        ],
                    )

                async with connection.cursor() as cursor:
                    sql = 'SELECT id, alpha, beta FROM foo ORDER BY id ASC'
                    await cursor.execute(sql)
                    self.assertEqual(
                        await cursor.fetchmany(-1),
                        [
                            (1, None, b'Beta test one'),
                            (2, 'Alpha test two', b'Beta test two'),
                        ],
                    )

                async with connection.cursor() as cursor:
                    sql = 'SELECT id, alpha, beta FROM foo ORDER BY id ASC'
                    await cursor.execute(sql)
                    self.assertEqual(
                        await cursor.fetchall(),
                        [
                            (1, None, b'Beta test one'),
                            (2, 'Alpha test two', b'Beta test two'),
                        ],
                    )

                async with connection.cursor() as cursor:
                    sql = 'SELECT id, alpha, beta FROM foo ORDER BY id ASC'
                    rows = []
                    async for row in await cursor.execute(sql):
                        rows.append(row)

                    self.assertEqual(
                        rows,
                        [
                            (1, None, b'Beta test one'),
                            (2, 'Alpha test two', b'Beta test two'),
                        ],
                    )
